
import std.stdio;
import std.format;
import std.string;

import core.sys.posix.signal;
import core.sys.posix.unistd;
import core.sys.posix.sys.ioctl;
import core.sys.linux.sys.signalfd;
import core.sys.linux.timerfd;
import core.stdc.errno;

import ecap, edsm;

class EventSource {
    EventQueue eq;
    int id;         /* fd */
    char tag;
    StageMachine owner;

    this(char tag) {
        this.tag = tag;
        eq = EventQueue.getInstance();
    }

    ~this() {
        writefln("   !!! %s() : %s (owner %s, fd = %d) this @ 0x%x", __FUNCTION__, this, owner.name, id, cast(void*)this);
    }

    void register() {
        // eq.registerEventSource(this);
        eq += this;
        writefln("'%s' registered %d (%s)", owner.name, id, this);
    }

    void deregister() {
        // eq.deregisterEventSource(this);
        eq -= this;
        writefln("'%s' deregistered %d (%s)", owner.name, id, this);
    }

    void enable() {
        eq.enableEventSource(this);
        writefln("'%s' enabled %d (%s)", owner.name, id, this);
    }

    void enableOnce() {
        eq.enableEventSourceOnce(this);
        writefln("'%s' enabled %d (%s, oneshot) ", owner.name, id, this);
    }

    void disable() {
        eq.disableEventSource(this);
        writefln("'%s' disabled %d (%s)", owner.name, id, this);
    }

    abstract ulong eventCode(int events);
}

final class Signal : EventSource {

    enum int sigInt = SIGINT;
    enum int sigTerm = SIGTERM;
    ulong number;

    this(int signo) {
        super('S');

        sigset_t sset;
        sigset_t old_sset;
        /* block the signal */
        sigemptyset(&sset);
        sigaddset(&sset, signo);
        sigprocmask(SIG_BLOCK, &sset, &old_sset);

        id = signalfd(-1, &sset, SFD_CLOEXEC);
        assert(id > 0, "signalfd() failed");
    }

    void flush() {
        signalfd_siginfo si;
        size_t r = read(id, &si, si.sizeof);
        assert(r == 128, "read(sigfd) failed");
        writefln("si.sizeof = %d", si.sizeof);
    }

    override ulong eventCode(int events) {
        flush();
        return number;
    }
}

final class Timer : EventSource {
    ulong nexp;     // >1 => overrun
    ulong number;

    this() {
        super('T');
        id = timerfd_create(CLOCK_REALTIME, 0);
        assert(id > 0, "timerfd_create() failed");
    }

    void heartBeat(uint interval_ms) {
        itimerspec t;

        /* first expiration */
        t.it_value.tv_sec = interval_ms / 1000;
        t.it_value.tv_nsec = (interval_ms % 1000) * 1000 * 1000;

        /* period */
        t.it_interval.tv_sec = t.it_value.tv_sec;
        t.it_interval.tv_nsec = t.it_value.tv_nsec;

        int r = timerfd_settime(id, 0, &t, null);
        assert(r == 0, "timerfd_settime() failed");
    }

    void oneShot(uint interval_ms) {
        itimerspec t;

        /* first expiration */
        t.it_value.tv_sec = interval_ms / 1000;
        t.it_value.tv_nsec = (interval_ms % 1000) * 1000 * 1000;

        /* period */
//        t.it_interval.tv_sec = 0;
//        t.it_interval.tv_nsec = 0;

        int r = timerfd_settime(id, 0, &t, null);
        assert(r == 0, "timerfd_settime() failed");
    }

    void stop() {
        itimerspec t;
        t.it_value.tv_sec = 0;
        t.it_value.tv_nsec = 0;
        t.it_interval.tv_sec = 0;
        t.it_interval.tv_nsec = 0;
        int r = timerfd_settime(id, 0, &t, null);
        assert(r == 0, "timerfd_settime() failed");
    }

    private void flush() {
        size_t r = read(id, &nexp, nexp.sizeof);
        assert(r == nexp.sizeof, "read(timerfd) failed");
        if (nexp > 1) {
            writefln("WRN: timer overrun! (number of expirations = %d)", nexp);
        }
    }

    override ulong eventCode(int events) {
        flush();
        return number;
    }
}

class Io : EventSource {

    this() {
        super('D');
        id = -1;
    }

    this(int fd) {
        super('D');
        id = fd;
    }

/*
    this(string serDev, int baudRate, string param = "8N1") {
        // TODO
        // int fd = open(serDev, ...);
        // setup...
        // this(fd);
    }
*/

    int bytesAvail() {
        int ba;
        int r = ioctl(id, FIONREAD, &ba);
        assert(r == 0, format("ioctl(%d) failed", id));
        return ba;
    }

    void enableCanWrite() {
        eq.enableEventSource(this, true);
        writefln("'%s' enabled %d (%s)", owner.name, id, this);
    }

    void enableCanWriteOnce() {
        eq.enableEventSourceOnce(this, true);
        writefln("'%s' enabled %d (%s, oneshot) ", owner.name, id, this);
    }

    override ulong eventCode(int events) {
        ulong ec;
        if (events & EventQueue.ePollIn) {
            ec = 0;
        }
        if (events & EventQueue.ePollOut) {
            ec = 1;
        }
        if (events & (EventQueue.ePollErr | EventQueue.ePollHup | EventQueue.ePollRdHup)) {
            ec = 2;
        }
        return ec;
    }
}

final class ClientSocket : Io {

    enum ConnectionState : int {
        NotConnected = 0,
        Connecting = 1,
        Connected = 2
    }

    import core.sys.posix.sys.socket;
    import core.sys.posix.netinet.in_;
    import core.sys.posix.netdb;
    import core.stdc.string;
    import core.sys.posix.fcntl;

    string host;
    string port;
    addrinfo hints;
    addrinfo *ai;

    ConnectionState state;

    this(string host, string port) {
        super();
        this.host = host;
        this.port = port;
        hints.ai_family = AF_INET;
        hints.ai_socktype = SOCK_STREAM;
    }

    void getId() {
        //getaddrinfo(host.ptr, port.ptr, &hints, &ai);
        getaddrinfo(toStringz(host), toStringz(port), &hints, &ai);
        id = socket(ai.ai_family, ai.ai_socktype, ai.ai_protocol);
        assert(id > 0, format("socket() failed"));
        state = ConnectionState.NotConnected;
    }

    void startConnect() {

        assert(state == ConnectionState.NotConnected, "");

        int flags = fcntl(id, F_GETFL);
        flags |= O_NONBLOCK;
        fcntl(id, F_SETFL, flags);

        int r = connect(id, ai.ai_addr, ai.ai_addrlen);
        if (-1 == r) {
            auto e = errno();
            assert(e == EINPROGRESS, "connect() failed");
        }
        // writefln("%s(): r = %d (%d)", __FUNCTION__, r, errno());
        state = ConnectionState.Connecting;
        // oops... and where to set to Connected?...
    }

    bool connOk() {
        /*
        tcp_info tcpi; // linux specific, see /usr/include/linux/tcp.h
        socklen_t tcpi_len = tcpi.sizeof;

        getsockopt(id, SOL_TCP, TCP_INFO, &tcpi, &tcpi_len);
        if (tcpi.tcpi_state != TCP_ESTABLISHED)
            return false;
        */
        return true; // TODO
    }

    void putId() {
        freeaddrinfo(ai);
        close(id);
    }

    int getError() {
        int e;
        socklen_t optlen = e.sizeof;
        getsockopt(id, SOL_SOCKET, SO_ERROR, &e, &optlen);
        return e;
    }
}

class Client {
    int fd;
    string addr;
    ushort port;

    this(int fd, string addr, ushort port) {
        this.fd = fd;
        this.addr = addr;
        this.port = port;
    }

    ~this() {
        writeln(__FUNCTION__);
        close(fd);
    }
}

final class TCPListener : EventSource {

    import core.sys.posix.sys.socket;
    import core.sys.posix.netinet.in_;
    import core.stdc.string;

    this(ushort port) {
        super('L');
        int sk = socket(AF_INET, SOCK_STREAM, 0);
        assert(sk > 0, format("socket() failed"));
        int on = 1;
        setsockopt(sk, SOL_SOCKET, SO_REUSEADDR, &on, on.sizeof);

        sockaddr_in myAddr;
        myAddr.sin_family = AF_INET;
        myAddr.sin_port = htons(port);
        bind(sk, cast(sockaddr*)&myAddr, myAddr.sizeof);
        listen(sk, 128);
        id = sk;
    }

    Client acceptClient() {
        sockaddr_in peerAddr;
        socklen_t addrLength = peerAddr.sizeof;
        char[32] addr;
        int sk = accept(id, cast(sockaddr*)&peerAddr, &addrLength);
        inet_ntop(AF_INET, &peerAddr.sin_addr, addr.ptr, cast(uint)addr.length);
        return new Client(sk, addr[0 .. strlen(addr.ptr)].idup, peerAddr.sin_port);
    }

    override ulong eventCode(int events) {
        ulong ec;
        if (events & EventQueue.ePollIn) {
            ec = 0;
        }
        if (events & (EventQueue.ePollErr | EventQueue.ePollHup | EventQueue.ePollRdHup)) {
            ec = 2;
        }
        return ec;
    }
}
